import pytest

PASSING_WORKFLOW = """\
name: build
on:
  push:
jobs:
  test:
    timeout-minutes: 60
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v2
      - uses: actions/setup-python@v2
      - name: install requirements
        run: python -m pip install tox
      - name: test
        run: python -m tox -e py
"""
FAILING_WORKFLOW = """\
name: build
on:
  push:
jobs:
  test:
    runs-on: ubuntu-latest
    steps:
      - uses: actions/checkout@v2
      - uses: actions/setup-python@v2
      - name: install requirements
        run: python -m pip install tox
      - name: test
        run: python -m tox -e py
"""


@pytest.mark.parametrize(
    "schemaname",
    ["github-workflows-require-timeout", "custom.github-workflows-require-timeout"],
)
@pytest.mark.parametrize(
    "vendor_schemaname",
    ["github-workflows", "vendor.github-workflows"],
)
def test_github_require_timeouts_passing(
    run_line_simple, tmp_path, schemaname, vendor_schemaname
):
    workflow = tmp_path / "doc.yaml"
    workflow.write_text(PASSING_WORKFLOW)

    # vendored github workflow schema passes on it
    run_line_simple(["--builtin-schema", vendor_schemaname, str(workflow)])

    run_line_simple(["--builtin-schema", schemaname, str(workflow)])


@pytest.mark.parametrize(
    "schemaname",
    ["github-workflows-require-timeout", "custom.github-workflows-require-timeout"],
)
@pytest.mark.parametrize(
    "vendor_schemaname",
    ["github-workflows", "vendor.github-workflows"],
)
def test_github_require_timeouts_failing(
    run_line, tmp_path, schemaname, vendor_schemaname
):
    workflow = tmp_path / "doc.yaml"
    workflow.write_text(FAILING_WORKFLOW)

    # vendored github workflow schema passes on it
    res1 = run_line(
        ["check-jsonschema", "--builtin-schema", vendor_schemaname, str(workflow)]
    )
    assert res1.exit_code == 0

    res2 = run_line(
        ["check-jsonschema", "--builtin-schema", schemaname, str(workflow)],
    )
    assert res2.exit_code == 1



from __future__ import annotations

import io
import typing as t

from check_jsonschema.cli.param_types import CustomLazyFile

from .parsers import ParseError, ParserSet
from .transforms import Transform


class InstanceLoader:
    def __init__(
        self,
        files: t.Sequence[t.IO[bytes] | CustomLazyFile],
        default_filetype: str = "json",
        data_transform: Transform | None = None,
    ) -> None:
        self._files = files
        self._default_filetype = default_filetype
        self._data_transform = (
            data_transform if data_transform is not None else Transform()
        )

        self._parsers = ParserSet(
            modify_yaml_implementation=self._data_transform.modify_yaml_implementation
        )

    def iter_files(self) -> t.Iterator[tuple[str, ParseError | t.Any]]:
        for file in self._files:
            if hasattr(file, "name"):
                name = file.name
            # allowing for BytesIO to be special-cased here is useful for
            # simpler test setup, since this is what tests will pass and we naturally
            # support it here
            elif isinstance(file, io.BytesIO) or file.fileno() == 0:
                name = "<stdin>"
            else:
                raise ValueError(f"File {file} has no name attribute")

            try:
                if isinstance(file, CustomLazyFile):
                    stream: t.IO[bytes] = t.cast(t.IO[bytes], file.open())
                else:
                    stream = file

                try:
                    data: t.Any = self._parsers.parse_data_with_path(
                        stream, name, self._default_filetype
                    )
                except ParseError as err:
                    data = err
                else:
                    data = self._data_transform(data)
            finally:
                file.close()
            yield (name, data)



from __future__ import annotations

import copy

import jsonschema
import jsonschema.validators

from ..regex_variants import RegexImplementation
from .implementations import validate_rfc3339, validate_time

# all known format strings except for a selection from draft3 which have either
# been renamed or removed:
# - color
# - host-name
# - ip-address
KNOWN_FORMATS: tuple[str, ...] = (
    "date",
    "date-time",
    "duration",
    "email",
    "hostname",
    "idn-email",
    "idn-hostname",
    "ipv4",
    "ipv6",
    "iri",
    "iri-reference",
    "json-pointer",
    "regex",
    "relative-json-pointer",
    "time",
    "uri",
    "uri-reference",
    "uri-template",
    "uuid",
)


class FormatOptions:
    def __init__(
        self,
        *,
        regex_impl: RegexImplementation,
        enabled: bool = True,
        disabled_formats: tuple[str, ...] = (),
    ) -> None:
        self.enabled = enabled
        self.regex_impl = regex_impl
        self.disabled_formats = disabled_formats


def get_base_format_checker(schema_dialect: str | None) -> jsonschema.FormatChecker:
    # resolve the dialect, if given, to a validator class
    # default to the latest draft
    validator_class = jsonschema.validators.validator_for(
        {} if schema_dialect is None else {"$schema": schema_dialect},
        default=jsonschema.Draft202012Validator,
    )
    return validator_class.FORMAT_CHECKER


def make_format_checker(
    opts: FormatOptions,
    schema_dialect: str | None = None,
) -> jsonschema.FormatChecker | None:
    if not opts.enabled:
        return None

    # customize around regex checking first
    checker = format_checker_for_regex_impl(opts.regex_impl)

    # add other custom format checks
    checker.checks("date-time")(validate_rfc3339)
    checker.checks("time")(validate_time)

    # remove the disabled checks, which may include the regex check
    for checkname in opts.disabled_formats:
        if checkname not in checker.checkers:
            continue
        del checker.checkers[checkname]

    return checker


def format_checker_for_regex_impl(
    regex_impl: RegexImplementation, schema_dialect: str | None = None
) -> jsonschema.FormatChecker:
    # convert to a schema-derived format checker, and copy it
    # for safe modification
    base_checker = get_base_format_checker(schema_dialect)
    checker = copy.deepcopy(base_checker)

    # replace the regex check
    del checker.checkers["regex"]
    checker.checks("regex")(regex_impl.check_format)

    return checker



from __future__ import annotations

import pathlib
import re

import pytest
import ruamel.yaml

yaml = ruamel.yaml.YAML(typ="safe")

HERE = pathlib.Path(__file__).parent


def check_pattern_match(
    pattern: str | re.Pattern, value: str, *, should_match: bool = True
) -> None:
    __tracebackhide__ = True
    if not isinstance(pattern, re.Pattern):
        pattern = re.compile(pattern)

    matches = bool(pattern.fullmatch(value))

    if matches == should_match:
        return

    if should_match:
        pytest.fail(f"'{pattern}' did not match '{value}' (expected match)")
    else:
        pytest.fail(f"'{pattern}' matched '{value}' (expected no match)")


def get_hook_config(hookid):
    config_file = HERE.parent.parent / ".pre-commit-hooks.yaml"
    with open(config_file) as fp:
        for hook in yaml.load(fp):
            if hook["id"] == hookid:
                return hook
        else:
            raise LookupError(f"could not find hook with id={hookid}")


_HOOKID_PATH_MAP = {
    "check-azure-pipelines": {
        "good": (
            "azure-pipelines.yml",
            "azure-pipelines.yaml",
            ".azure-pipelines.yml",
            ".azure-pipelines.yaml",
        ),
        "bad": (
            "foo.yml",
            "foo/azure-pipelines.yaml",
        ),
    },
    "check-bamboo-spec": {
        "good": (
            "bamboo-specs/foo.yml",
            "bamboo-specs/foo.yaml",
        ),
        "bad": (
            "bamboo-specs.yaml",
            "bamboo-spec/foo.yml",
            "bamboo-specs/README.md",
        ),
    },
    "check-compose-spec": {
        "good": (
            "compose.yml",
            "compose.yaml",
            "docker-compose.yml",
            "docker-compose.yaml",
            "compose.override.yml",
            "docker-compose.override.yml",
            "path/to/compose.yml",
        ),
        "bad": (
            "docker.compose.yml",
            "docker.md",
            "Dockerfile",
        ),
    },
    "check-meltano": {
        "good": (
            "meltano.yml",
            "data/meltano.yml",
            "extractors.meltano.yml",
            "meltano-manifest.json",
            "meltano-manifest.prod.json",
        ),
        "bad": (
            "meltano.yaml",
            "meltano.yml.md",
            "meltano-manifest.yml",
        ),
    },
    "check-dependabot": {
        "good": (".github/dependabot.yml", ".github/dependabot.yaml"),
        "bad": (".dependabot.yaml", ".dependabot.yml"),
    },
    "check-github-actions": {
        "good": (
            "action.yaml",
            ".github/actions/action.yml",
            ".github/actions/foo/bar/action.yaml",
            ".github/actions/path with spaces/action.yml",
        ),
        "bad": (".github/actions/foo/other.yaml",),
    },
    "check-github-workflows": {
        "good": (
            ".github/workflows/build.yml",
            ".github/workflows/build.yaml",
        ),
        "bad": (
            ".github/workflows.yaml",
            ".github/workflows/foo/bar.yaml",
        ),
    },
    "check-gitlab-ci": {
        "good": (
            ".gitlab-ci.yml",
            ".gitlab/.gitlab-ci.yml",
            "gitlab/.gitlab-ci.yml",
        ),
        "bad": (
            ".gitlab-ci.yaml",
            "gitlab-ci.yml",
            ".gitlab/gitlab-ci.yml",
            "gitlab/gitlab-ci.yml",
        ),
    },
    "check-readthedocs": {
        "good": (
            ".readthedocs.yml",
            ".readthedocs.yaml",
        ),
        "bad": (
            "readthedocs.yml",
            "readthedocs.yaml",
        ),
    },
    "check-renovate": {
        "good": (
            "renovate.json",
            "renovate.json5",
            ".github/renovate.json",
            ".gitlab/renovate.json5",
            ".renovaterc",
            ".renovaterc.json",
        ),
        "bad": (
            ".github/renovaterc",
            ".renovate",
            ".renovate.json",
        ),
    },
    "check-snapcraft": {
        "good": ("snapcraft.yaml", "snap/snapcraft.yaml", "foo/bar/snapcraft.yaml"),
        "bad": ("snapcraft.yml", "snap.yaml", "snapcraft"),
    },
    "check-travis": {
        "good": (
            ".travis.yml",
            ".travis.yaml",
        ),
        "bad": (
            "travis.yml",
            ".travis",
        ),
    },
}


@pytest.mark.parametrize(
    "hookid, filepath",
    [
        (hookid, path)
        for (hookid, pathlist) in _HOOKID_PATH_MAP.items()
        for path in pathlist.get("good", ())
    ],
)
def test_hook_matches_known_good_paths(hookid, filepath):
    config = get_hook_config(hookid)
    check_pattern_match(config["files"], filepath)


@pytest.mark.parametrize(
    "hookid, filepath",
    [
        (hookid, path)
        for (hookid, pathlist) in _HOOKID_PATH_MAP.items()
        for path in pathlist.get("bad", ())
    ],
)
def test_hook_does_not_matches_known_bad_paths(hookid, filepath):
    config = get_hook_config(hookid)
    check_pattern_match(config["files"], filepath, should_match=False)



import pytest

from check_jsonschema.instance_loader import InstanceLoader
from check_jsonschema.parsers import BadFileTypeError, FailedFileLoadError
from check_jsonschema.parsers.json5 import ENABLED as JSON5_ENABLED


# handy helper for opening multiple files for InstanceLoader
@pytest.fixture
def open_wide():
    track_paths = []

    def func(*paths):
        open_paths = [open(p, "rb") for p in paths]
        track_paths.extend(open_paths)
        return open_paths

    yield func

    for p in track_paths:
        p.close()


@pytest.mark.parametrize(
    "filename, default_filetype",
    [
        ("foo.json", "notarealfiletype"),
        ("foo.json", "json"),
        ("foo.json", "yaml"),
        ("foo", "json"),
        # YAML is a superset of JSON, so using the YAML loader should be safe when the
        # data is JSON
        ("foo", "yaml"),
    ],
)
def test_instanceloader_json_data(tmp_path, filename, default_filetype, open_wide):
    f = tmp_path / filename
    f.write_text("{}")
    loader = InstanceLoader(open_wide(f), default_filetype=default_filetype)
    data = list(loader.iter_files())
    assert data == [(str(f), {})]


@pytest.mark.parametrize(
    "filename, default_filetype",
    [
        ("foo.yaml", "notarealfiletype"),
        ("foo.yml", "notarealfiletype"),
        ("foo.yaml", "json"),
        ("foo.yml", "json"),
        ("foo.yaml", "yaml"),
        ("foo.yml", "yaml"),
        ("foo", "yaml"),
    ],
)
def test_instanceloader_yaml_data(tmp_path, filename, default_filetype, open_wide):
    f = tmp_path / filename
    f.write_text(
        """\
a:
  b:
   - 1
   - 2
  c: d
"""
    )
    loader = InstanceLoader(open_wide(f), default_filetype=default_filetype)
    data = list(loader.iter_files())
    assert data == [(str(f), {"a": {"b": [1, 2], "c": "d"}})]


@pytest.mark.parametrize(
    "filename, default_filetype",
    [
        ("foo.toml", "notarealfiletype"),
        ("foo.toml", "json"),
        ("foo.toml", "yaml"),
        ("foo", "toml"),
    ],
)
def test_instanceloader_toml_data(tmp_path, filename, default_filetype, open_wide):
    f = tmp_path / "foo.toml"
    f.write_text('[foo]\nbar = "baz"\n')
    loader = InstanceLoader(open_wide(f), default_filetype=default_filetype)
    data = list(loader.iter_files())
    assert data == [(str(f), {"foo": {"bar": "baz"}})]


def test_instanceloader_unknown_type_nonjson_content(tmp_path, open_wide):
    f = tmp_path / "foo"  # no extension here
    f.write_text("a:b")  # non-json data (cannot be detected as JSON)
    loader = InstanceLoader(open_wide(f), default_filetype="unknown")
    # at iteration time, the file should error and be reported as such
    data = list(loader.iter_files())
    assert len(data) == 1
    assert isinstance(data[0], tuple)
    assert len(data[0]) == 2
    assert data[0][0] == str(f)
    assert isinstance(data[0][1], BadFileTypeError)


@pytest.mark.parametrize(
    "enabled_flag, extension, file_content, expect_data, expect_error_message",
    [
        (
            JSON5_ENABLED,
            "json5",
            "{}",
            {},
            "pip install json5",
        ),
    ],
)
def test_instanceloader_optional_format_handling(
    tmp_path,
    enabled_flag,
    extension,
    file_content,
    expect_data,
    expect_error_message,
    open_wide,
):
    f = tmp_path / f"foo.{extension}"
    f.write_text(file_content)
    loader = InstanceLoader(open_wide(f))
    if enabled_flag:
        # at iteration time, the file should load fine
        data = list(loader.iter_files())
        assert data == [(str(f), expect_data)]
    else:
        # at iteration time, an error should be raised
        data = list(loader.iter_files())
        assert len(data) == 1
        assert isinstance(data[0], tuple)
        assert len(data[0]) == 2
        assert data[0][0] == str(f)
        assert isinstance(data[0][1], BadFileTypeError)

        # error message should be instructive
        assert expect_error_message in str(data[0])


def test_instanceloader_yaml_dup_anchor(tmp_path, open_wide):
    f = tmp_path / "foo.yaml"
    f.write_text(
        """\
a:
  b: &anchor
   - 1
   - 2
  c: &anchor d
"""
    )
    loader = InstanceLoader(open_wide(f))
    data = list(loader.iter_files())
    assert data == [(str(f), {"a": {"b": [1, 2], "c": "d"}})]


@pytest.mark.parametrize(
    "file_format, filename, content",
    [
        ("json", "foo.json", '{"a":\n'),
        ("yaml", "foo.yaml", "a: {b\n"),
        ("yaml", "foo.yaml", "a: b\nc\n"),
        ("json5", "foo.json5", '{"a":\n'),
        ("toml", "foo.toml", "abc\n"),
    ],
)
def test_instanceloader_invalid_data(
    tmp_path, file_format, filename, content, open_wide
):
    if file_format == "json5" and not JSON5_ENABLED:
        pytest.skip("test requires 'json5' support")

    f = tmp_path / filename
    f.write_text(content)
    loader = InstanceLoader(open_wide(f))
    data = list(loader.iter_files())
    assert len(data) == 1
    assert isinstance(data[0], tuple)
    assert len(data[0]) == 2
    assert data[0][0] == str(f)
    assert isinstance(data[0][1], FailedFileLoadError)


def test_instanceloader_invalid_data_mixed_with_valid_data(tmp_path, open_wide):
    a = tmp_path / "a.json"
    b = tmp_path / "b.json"
    c = tmp_path / "c.json"
    a.write_text("{}")
    b.write_text("{")
    c.write_text('{"c":true}')

    loader = InstanceLoader(open_wide(a, b, c))

    data = list(loader.iter_files())
    assert len(data) == 3

    assert data[0] == (str(a), {})

    assert isinstance(data[1], tuple)
    assert len(data[1]) == 2
    assert data[1][0] == str(b)
    assert isinstance(data[1][1], FailedFileLoadError)

    assert data[2] == (str(c), {"c": True})


@pytest.mark.parametrize(
    "filetypes",
    (
        ("json", "yaml"),
        ("json", "json5"),
        ("yaml", "toml"),
        ("json", "yaml", "toml", "json5"),
    ),
)
def test_instanceloader_mixed_filetypes(tmp_path, filetypes, open_wide):
    if not JSON5_ENABLED and "json5" in filetypes:
        pytest.skip("test requires json5")
    files = {}
    file_order = []
    if "json" in filetypes:
        files["json"] = tmp_path / "F.json"
        files["json"].write_text("{}")
        file_order.append("json")
    if "yaml" in filetypes:
        files["yaml"] = tmp_path / "F.yaml"
        files["yaml"].write_text("foo: bar")
        file_order.append("yaml")
    if "json5" in filetypes:
        files["json5"] = tmp_path / "F.json5"
        files["json5"].write_text('{//hi\n"c": 1}')
        file_order.append("json5")
    if "toml" in filetypes:
        files["toml"] = tmp_path / "F.toml"
        files["toml"].write_text('[foo]  # bar\nname = "value"\n')
        file_order.append("toml")

    loader = InstanceLoader(open_wide(*files.values()))

    data = list(loader.iter_files())
    assert len(data) == len(files)

    for i, filetype in enumerate(file_order):
        assert isinstance(data[i], tuple)
        assert len(data[i]) == 2
        path, value = data[i]
        assert path == str(files[filetype])
        if filetype == "json":
            assert value == {}
        elif filetype == "yaml":
            assert value == {"foo": "bar"}
        elif filetype == "json5":
            assert value == {"c": 1}
        elif filetype == "toml":
            assert value == {"foo": {"name": "value"}}



from __future__ import annotations

import enum
import typing as t

import click
import jsonschema

from ..formats import FormatOptions
from ..regex_variants import RegexImplementation, RegexVariantName
from ..transforms import Transform


class SchemaLoadingMode(enum.Enum):
    filepath = "filepath"
    builtin = "builtin"
    metaschema = "metaschema"


class ParseResult:
    def __init__(self) -> None:
        # primary options: schema + instances
        self.schema_mode: SchemaLoadingMode = SchemaLoadingMode.filepath
        self.schema_path: str | None = None
        self.base_uri: str | None = None
        self.instancefiles: tuple[t.IO[bytes], ...] = ()
        # cache controls
        self.disable_cache: bool = False
        self.cache_filename: str | None = None
        # filetype detection (JSON, YAML, TOML, etc)
        self.default_filetype: str = "json"
        # data-transform (for Azure Pipelines and potentially future transforms)
        self.data_transform: Transform | None = None
        # validation behavioral controls
        self.validator_class: type[jsonschema.protocols.Validator] | None = None
        self.fill_defaults: bool = False
        # regex format options
        self.disable_all_formats: bool = False
        self.disable_formats: tuple[str, ...] = ()
        self.regex_variant: RegexVariantName = RegexVariantName.default
        # error and output controls
        self.verbosity: int = 1
        self.traceback_mode: str = "short"
        self.output_format: str = "text"

    def set_regex_variant(
        self,
        variant_opt: t.Literal["python", "nonunicode", "default"] | None,
        *,
        legacy_opt: t.Literal["python", "nonunicode", "default"] | None = None,
    ) -> None:
        variant_name: t.Literal["python", "nonunicode", "default"] | None = (
            variant_opt or legacy_opt
        )
        if variant_name:
            self.regex_variant = RegexVariantName(variant_name)

    def set_schema(
        self, schemafile: str | None, builtin_schema: str | None, check_metaschema: bool
    ) -> None:
        mutex_arg_count = sum(
            1 if x else 0 for x in (schemafile, builtin_schema, check_metaschema)
        )
        if mutex_arg_count == 0:
            raise click.UsageError(
                "Either --schemafile, --builtin-schema, or --check-metaschema "
                "must be provided"
            )
        if mutex_arg_count > 1:
            raise click.UsageError(
                "--schemafile, --builtin-schema, and --check-metaschema "
                "are mutually exclusive"
            )

        if schemafile:
            self.schema_mode = SchemaLoadingMode.filepath
            self.schema_path = schemafile
        elif builtin_schema:
            self.schema_mode = SchemaLoadingMode.builtin
            self.schema_path = builtin_schema
        else:
            self.schema_mode = SchemaLoadingMode.metaschema

    def set_validator(
        self, validator_class: type[jsonschema.protocols.Validator] | None
    ) -> None:
        if validator_class is None:
            return
        if self.schema_mode != SchemaLoadingMode.filepath:
            raise click.UsageError(
                "--validator-class can only be used with --schemafile for schema loading"
            )
        self.validator_class = validator_class

    @property
    def format_opts(self) -> FormatOptions:
        return FormatOptions(
            regex_impl=RegexImplementation(self.regex_variant),
            enabled=not self.disable_all_formats,
            disabled_formats=self.disable_formats,
        )



import enum
import re
import typing as t

import jsonschema
import regress


class RegexVariantName(enum.Enum):
    default = "default"
    nonunicode = "nonunicode"
    python = "python"


class RegexImplementation:
    """
    A high-level interface for getting at the different possible
    implementations of regex behaviors.
    """

    _concrete: "_ConcreteImplementation"

    def __init__(self, variant: RegexVariantName) -> None:
        self.variant = variant

        if self.variant == RegexVariantName.default:
            self._concrete = _RegressImplementation()
        elif self.variant == RegexVariantName.nonunicode:
            self._concrete = _NonunicodeRegressImplementation()
        else:
            self._concrete = _PythonImplementation()

        self.check_format = self._concrete.check_format
        self.pattern_keyword = self._concrete.pattern_keyword
        self.patternProperties_keyword = self._concrete.patternProperties_keyword


class _ConcreteImplementation(t.Protocol):
    def check_format(self, instance: t.Any) -> bool: ...

    def pattern_keyword(
        self, validator: t.Any, pattern: str, instance: str, schema: t.Any
    ) -> t.Iterator[jsonschema.ValidationError]: ...

    def patternProperties_keyword(
        self,
        validator: t.Any,
        patternProperties: dict[str, t.Any],
        instance: dict[str, t.Any],
        schema: t.Any,
    ) -> t.Iterator[jsonschema.ValidationError]: ...


class _RegressImplementation:
    def _compile_pattern(self, pattern: str) -> regress.Regex:
        return regress.Regex(pattern, flags="u")

    def check_format(self, instance: t.Any) -> bool:
        if not isinstance(instance, str):
            return True
        try:
            self._compile_pattern(instance)
        except regress.RegressError:
            return False
        return True

    def pattern_keyword(
        self, validator: t.Any, pattern: str, instance: str, schema: t.Any
    ) -> t.Iterator[jsonschema.ValidationError]:
        if not validator.is_type(instance, "string"):
            return

        regress_pattern = self._compile_pattern(pattern)
        if not regress_pattern.find(instance):
            yield jsonschema.ValidationError(f"{instance!r} does not match {pattern!r}")

    def patternProperties_keyword(
        self,
        validator: t.Any,
        patternProperties: dict[str, t.Any],
        instance: dict[str, t.Any],
        schema: t.Any,
    ) -> t.Iterator[jsonschema.ValidationError]:
        if not validator.is_type(instance, "object"):
            return

        for pattern, subschema in patternProperties.items():
            regress_pattern = self._compile_pattern(pattern)
            for k, v in instance.items():
                if regress_pattern.find(k):
                    yield from validator.descend(
                        v,
                        subschema,
                        path=k,
                        schema_path=pattern,
                    )


class _NonunicodeRegressImplementation(_RegressImplementation):
    def _compile_pattern(self, pattern: str) -> regress.Regex:
        return regress.Regex(pattern)


class _PythonImplementation:
    def check_format(self, instance: t.Any) -> bool:
        if not isinstance(instance, str):
            return True
        try:
            re.compile(instance)
        except re.error:
            return False
        return True

    def pattern_keyword(
        self, validator: t.Any, pattern: str, instance: str, schema: t.Any
    ) -> t.Iterator[jsonschema.ValidationError]:
        if not validator.is_type(instance, "string"):
            return

        re_pattern = re.compile(pattern)
        if not re_pattern.search(instance):
            yield jsonschema.ValidationError(f"{instance!r} does not match {pattern!r}")

    def patternProperties_keyword(
        self,
        validator: t.Any,
        patternProperties: dict[str, t.Any],
        instance: dict[str, t.Any],
        schema: t.Any,
    ) -> t.Iterator[jsonschema.ValidationError]:
        if not validator.is_type(instance, "object"):
            return

        for pattern, subschema in patternProperties.items():
            for k, v in instance.items():
                if re.search(pattern, k):
                    yield from validator.descend(
                        v,
                        subschema,
                        path=k,
                        schema_path=pattern,
                    )



from pathlib import Path

import ruamel.yaml

from check_jsonschema.catalog import SCHEMA_CATALOG

yaml = ruamel.yaml.YAML(typ="safe")
HERE = Path(__file__).parent
CONFIG_FILE = HERE.parent.parent / ".pre-commit-hooks.yaml"


def test_schema_catalog_is_alphabetized():
    catalog_keys = list(SCHEMA_CATALOG.keys())
    sorted_keys = sorted(catalog_keys)
    assert catalog_keys == sorted_keys


def test_hooks_cover_catalog():
    with open(CONFIG_FILE, "rb") as fp:
        config = yaml.load(fp)
    config_hook_ids = {x["id"] for x in config}
    catalog_hook_ids = {f"check-{name}" for name in SCHEMA_CATALOG}
    assert catalog_hook_ids <= config_hook_ids



# test on a JavaScript regex which is not a valid python regex
# `--regex-variant=default` should accept it
# `--regex-variant=python` should reject it
#
# check these options against documents with invalid and valid python regexes to confirm
# that they are behaving as expected
import json

import pytest

FORMAT_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema",
    "properties": {"pattern": {"type": "string", "format": "regex"}},
}

ALWAYS_PASSING_DOCUMENT = {
    "pattern": "ab*c",
}

ALWAYS_FAILING_DOCUMENT = {
    "pattern": "a(b*c",
}

JS_REGEX_DOCUMENT = {
    "pattern": "a(?<captured>)bc",
}

# taken from https://github.com/python-jsonschema/check-jsonschema/issues/25
RENOVATE_DOCUMENT = {
    "regexManagers": [
        {
            "fileMatch": ["^Dockerfile$"],
            "matchStrings": ["ENV YARN_VERSION=(?<currentValue>.*?)\n"],
            "depNameTemplate": "yarn",
            "datasourceTemplate": "npm",
        }
    ]
}


@pytest.fixture(
    params=[
        ("--disable-formats", "regex"),
        ("--format-regex", "default"),
        ("--format-regex", "python"),
        ("--regex-variant", "python"),
        ("--regex-variant", "default"),
        ("--regex-variant", "default", "--format-regex", "python"),
        ("--regex-variant", "python", "--format-regex", "default"),
    ]
)
def regexopts(request):
    return request.param


def test_regex_format_good(run_line_simple, tmp_path, regexopts):
    schemafile = tmp_path / "schema.json"
    schemafile.write_text(json.dumps(FORMAT_SCHEMA))

    doc = tmp_path / "doc.json"
    doc.write_text(json.dumps(ALWAYS_PASSING_DOCUMENT))

    run_line_simple([*regexopts, "--schemafile", str(schemafile), str(doc)])


def test_regex_format_accepts_non_str_inputs(run_line_simple, tmp_path, regexopts):
    # potentially confusing, but a format checker is allowed to check non-str instances
    # validate the format checker behavior on such a case
    schemafile = tmp_path / "schema.json"
    schemafile.write_text(
        json.dumps(
            {
                "$schema": "http://json-schema.org/draft-07/schema",
                "properties": {"pattern": {"type": "integer", "format": "regex"}},
            }
        )
    )
    doc = tmp_path / "doc.json"
    doc.write_text(json.dumps({"pattern": 0}))
    run_line_simple([*regexopts, "--schemafile", str(schemafile), str(doc)])


def test_regex_format_bad(run_line, tmp_path, regexopts):
    schemafile = tmp_path / "schema.json"
    schemafile.write_text(json.dumps(FORMAT_SCHEMA))

    doc = tmp_path / "doc.json"
    doc.write_text(json.dumps(ALWAYS_FAILING_DOCUMENT))

    expect_ok = regexopts == ("--disable-formats", "regex")

    res = run_line(
        [
            "check-jsonschema",
            *regexopts,
            "--schemafile",
            str(schemafile),
            str(doc),
        ],
    )
    if expect_ok:
        assert res.exit_code == 0
    else:
        assert res.exit_code == 1
        assert "is not a 'regex'" in res.stdout


def test_regex_format_js_specific(run_line, tmp_path, regexopts):
    schemafile = tmp_path / "schema.json"
    schemafile.write_text(json.dumps(FORMAT_SCHEMA))

    doc = tmp_path / "doc.json"
    doc.write_text(json.dumps(JS_REGEX_DOCUMENT))

    expect_ok = regexopts[:2] not in (
        ("--format-regex", "python"),
        ("--regex-variant", "python"),
    )

    res = run_line(
        [
            "check-jsonschema",
            *regexopts,
            "--schemafile",
            str(schemafile),
            str(doc),
        ],
    )
    if expect_ok:
        assert res.exit_code == 0
    else:
        assert res.exit_code == 1
        assert "is not a 'regex'" in res.stdout


def test_regex_format_in_renovate_config(run_line_simple, tmp_path):
    doc = tmp_path / "doc.json"
    doc.write_text(json.dumps(RENOVATE_DOCUMENT))

    run_line_simple(["--builtin-schema", "vendor.renovate", str(doc)])



import re

TIME_REGEX = re.compile(
    r"""
    ^
    (?:[01]\d|2[0123])
    :
    (?:[0-5]\d)
    :
    (?:[0-5]\d)
    # (optional) fractional seconds
    (?:(\.|,)\d+)?
    # UTC or offset
    (?:
        Z
        | z
        | [+-](?:[01]\d|2[0123]):[0-5]\d
    )
    $
""",
    re.VERBOSE | re.ASCII,
)


def validate(time_str: object) -> bool:
    if not isinstance(time_str, str):
        return False
    return bool(TIME_REGEX.match(time_str))


if __name__ == "__main__":
    import timeit

    N = 100_000
    tests = (
        ("basic", "23:59:59Z"),
        ("long_fracsec", "23:59:59.8446519776713Z"),
    )

    print("benchmarking")
    for name, val in tests:
        all_times = timeit.repeat(
            f"validate({val!r})", globals=globals(), repeat=3, number=N
        )
        print(f"{name} (valid={validate(val)}): {int(min(all_times) / N * 10**9)}ns")



import pytest

from check_jsonschema.parsers.yaml import ParseError, construct_yaml_implementation
from check_jsonschema.transforms.gitlab import (
    GITLAB_TRANSFORM,
    GitLabReferenceExpectationViolation,
)


def test_can_parse_yaml_with_transform():
    rawdata = """\
a: b
c: d
"""

    impl = construct_yaml_implementation()

    data = impl.load(rawdata)
    assert data == {"a": "b", "c": "d"}

    GITLAB_TRANSFORM.modify_yaml_implementation(impl)
    data = impl.load(rawdata)
    assert data == {"a": "b", "c": "d"}


def test_can_parse_ok_gitlab_yaml_with_transform():
    rawdata = """\
foo:
- !reference [bar, baz]
"""
    impl = construct_yaml_implementation()

    with pytest.raises(ParseError):
        data = impl.load(rawdata)

    GITLAB_TRANSFORM.modify_yaml_implementation(impl)
    data = impl.load(rawdata)
    assert data == {"foo": [["bar", "baz"]]}


def test_cannot_parse_bad_gitlab_yaml_with_transform():
    rawdata = """\
foo:
- !reference true
"""
    impl = construct_yaml_implementation()

    with pytest.raises(ParseError):
        impl.load(rawdata)

    GITLAB_TRANSFORM.modify_yaml_implementation(impl)
    with pytest.raises(
        GitLabReferenceExpectationViolation,
        match=r"check-jsonschema rejects this gitlab \!reference tag: .*",
    ):
        impl.load(rawdata)



from __future__ import annotations

import typing as t
import urllib.parse

import referencing
from referencing.jsonschema import DRAFT202012, Schema

from ..cachedownloader import CacheDownloader
from ..parsers import ParserSet
from ..utils import filename2path


def make_reference_registry(
    parsers: ParserSet, retrieval_uri: str | None, schema: dict, disable_cache: bool
) -> referencing.Registry:
    id_attribute_: t.Any = schema.get("$id")
    if isinstance(id_attribute_, str):
        id_attribute: str | None = id_attribute_
    else:
        id_attribute = None

    schema_resource = referencing.Resource.from_contents(
        schema, default_specification=DRAFT202012
    )
    # mypy does not recognize that Registry is an `attrs` class and has `retrieve` as an
    # argument to its implicit initializer
    registry: referencing.Registry = referencing.Registry(  # type: ignore[call-arg]
        retrieve=create_retrieve_callable(
            parsers, retrieval_uri, id_attribute, disable_cache
        )
    )

    if retrieval_uri is not None:
        registry = registry.with_resource(uri=retrieval_uri, resource=schema_resource)
    if id_attribute is not None:
        registry = registry.with_resource(uri=id_attribute, resource=schema_resource)

    return registry


def create_retrieve_callable(
    parser_set: ParserSet,
    retrieval_uri: str | None,
    id_attribute: str | None,
    disable_cache: bool,
) -> t.Callable[[str], referencing.Resource[Schema]]:
    base_uri = id_attribute
    if base_uri is None:
        base_uri = retrieval_uri

    cache = ResourceCache()
    downloader = CacheDownloader("refs", disable_cache=disable_cache)

    def get_local_file(uri: str) -> t.Any:
        path = filename2path(uri)
        return parser_set.parse_file(path, "json")

    def retrieve_reference(uri: str) -> referencing.Resource[Schema]:
        scheme = urllib.parse.urlsplit(uri).scheme
        if scheme == "" and base_uri is not None:
            full_uri = urllib.parse.urljoin(base_uri, uri)
        else:
            full_uri = uri

        if full_uri in cache:
            return cache[full_uri]

        full_uri_scheme = urllib.parse.urlsplit(full_uri).scheme
        if full_uri_scheme in ("http", "https"):

            def validation_callback(content: bytes) -> None:
                parser_set.parse_data_with_path(content, full_uri, "json")

            bound_downloader = downloader.bind(
                full_uri, validation_callback=validation_callback
            )
            with bound_downloader.open() as fp:
                data = fp.read()

            parsed_object = parser_set.parse_data_with_path(data, full_uri, "json")
        else:
            parsed_object = get_local_file(full_uri)

        cache[full_uri] = parsed_object
        return cache[full_uri]

    return retrieve_reference


class ResourceCache:
    def __init__(self) -> None:
        self._cache: t.Dict[str, referencing.Resource[Schema]] = {}

    def __setitem__(self, uri: str, data: t.Any) -> referencing.Resource[Schema]:
        resource = referencing.Resource.from_contents(
            data, default_specification=DRAFT202012
        )
        self._cache[uri] = resource
        return resource

    def __getitem__(self, uri: str) -> referencing.Resource[Schema]:
        return self._cache[uri]

    def __contains__(self, uri: str) -> bool:
        return uri in self._cache



import textwrap

import pytest
from click.testing import CliRunner

from check_jsonschema import main as cli_main


def _render_result(result):
    return f"""
output:
{textwrap.indent(result.output, "  ")}

stderr:
{textwrap.indent(result.stderr, "  ")}
"""


@pytest.fixture
def cli_runner():
    return CliRunner(mix_stderr=False)


@pytest.fixture
def run_line(cli_runner):
    def func(cli_args, *args, **kwargs):
        assert cli_args[0] == "check-jsonschema"
        if "catch_exceptions" not in kwargs:
            kwargs["catch_exceptions"] = False
        return cli_runner.invoke(cli_main, cli_args[1:], *args, **kwargs)

    return func


@pytest.fixture
def run_line_simple(run_line):
    def func(cli_args, *args, full_traceback: bool = True, **kwargs):
        res = run_line(
            ["check-jsonschema"]
            + (["--traceback-mode", "full"] if full_traceback else [])
            + cli_args,
            *args,
            **kwargs,
        )
        assert res.exit_code == 0, _render_result(res)

    return func
